package dev.shadowsoffire.apotheosis.adventure.compat;

import twilightforest.capabilities.shield.IShieldCapability;
import twilightforest.capabilities.CapabilityList;
import java.util.function.Function;
import net.minecraft.network.chat.Style;
import net.minecraft.util.StringUtil;
import net.minecraft.world.effect.MobEffectInstance;
import net.minecraft.world.effect.MobEffects;
import net.minecraft.util.Mth;
import net.minecraft.world.entity.ai.attributes.AttributeModifier;
import net.minecraft.world.entity.ai.attributes.Attributes;
import net.minecraft.network.chat.TextColor;
import dev.shadowsoffire.placebo.color.GradientColor;
import net.minecraft.nbt.CompoundTag;
import dev.shadowsoffire.apotheosis.adventure.affix.Affix;
import java.util.function.BiFunction;
import com.mojang.datafixers.kinds.Applicative;
import com.mojang.datafixers.kinds.App;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import net.minecraft.ChatFormatting;
import net.minecraft.network.chat.Component;
import com.google.common.base.Preconditions;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.InteractionResult;
import net.minecraft.world.item.context.UseOnContext;
import net.minecraft.world.item.ItemStack;
import dev.shadowsoffire.apotheosis.adventure.socket.gem.GemClass;
import dev.shadowsoffire.placebo.util.StepFunction;
import dev.shadowsoffire.apotheosis.adventure.loot.LootRarity;
import java.util.Map;
import net.minecraft.core.registries.Registries;
import net.minecraft.resources.ResourceLocation;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.ai.goal.Goal;
import net.minecraft.world.entity.PathfinderMob;
import net.minecraft.world.entity.ai.goal.AvoidEntityGoal;
import net.minecraft.world.entity.player.Player;
import java.util.function.Predicate;
import com.google.common.base.Predicates;
import net.minecraftforge.event.entity.EntityJoinLevelEvent;
import java.util.function.Consumer;
import net.minecraftforge.common.MinecraftForge;
import com.mojang.serialization.Codec;
import dev.shadowsoffire.apotheosis.Apotheosis;
import dev.shadowsoffire.apotheosis.adventure.socket.gem.bonus.GemBonus;
import twilightforest.entity.monster.Redcap;
import net.minecraft.world.entity.EntityType;
import net.minecraft.world.item.Item;
import net.minecraftforge.registries.RegistryObject;

public class AdventureTwilightCompat
{
    protected static final RegistryObject<Item> ORE_MAGNET;
    protected static final RegistryObject<EntityType<Redcap>> REDCAP;
    
    public static void register() {
        GemBonus.CODEC.register(Apotheosis.loc("twilight_ore_magnet"), (Codec)OreMagnetBonus.CODEC);
        GemBonus.CODEC.register(Apotheosis.loc("twilight_treasure_goblin"), (Codec)TreasureGoblinBonus.CODEC);
        GemBonus.CODEC.register(Apotheosis.loc("twilight_fortification"), (Codec)FortificationBonus.CODEC);
        MinecraftForge.EVENT_BUS.addListener((Consumer)AdventureTwilightCompat::doGoblins);
    }
    
    @SubscribeEvent
    public static void doGoblins(final EntityJoinLevelEvent e) {
        final Entity entity = e.getEntity();
        if (entity instanceof final Redcap r) {
            if (r.getPersistentData().m_128441_("apoth.treasure_goblin")) {
                r.f_21346_.m_262460_((Predicate)Predicates.alwaysTrue());
                r.f_21345_.m_262460_((Predicate)Predicates.alwaysTrue());
                r.f_21345_.m_25352_(10, (Goal)new AvoidEntityGoal((PathfinderMob)r, (Class)Player.class, 6.0f, 1.0, 1.25));
            }
        }
    }
    
    static {
        ORE_MAGNET = RegistryObject.create(new ResourceLocation("twilightforest", "ore_magnet"), Registries.f_256913_, "apotheosis");
        REDCAP = RegistryObject.create(new ResourceLocation("twilightforest", "redcap"), Registries.f_256939_, "apotheosis");
    }
    
    public static class OreMagnetBonus extends GemBonus
    {
        public static final Codec<OreMagnetBonus> CODEC;
        protected final Map<LootRarity, StepFunction> values;
        
        public OreMagnetBonus(final GemClass gemClass, final Map<LootRarity, StepFunction> values) {
            super(Apotheosis.loc("twilight_ore_magnet"), gemClass);
            this.values = values;
        }
        
        public InteractionResult onItemUse(final ItemStack gem, final LootRarity rarity, final UseOnContext ctx) {
            final BlockState state = ctx.m_43725_().m_8055_(ctx.m_8083_());
            if (state.m_60795_()) {
                return null;
            }
            final Level level = ctx.m_43725_();
            final Player player = ctx.m_43723_();
            player.m_6672_(ctx.m_43724_());
            ((Item)AdventureTwilightCompat.ORE_MAGNET.get()).m_5551_(gem, level, (LivingEntity)player, 0);
            player.m_5810_();
            final int cost = this.values.get(rarity).getInt(0.0f);
            ctx.m_43722_().m_41622_(cost, (LivingEntity)player, user -> user.m_21190_(ctx.m_43724_()));
            return super.onItemUse(gem, rarity, ctx);
        }
        
        public Codec<? extends GemBonus> getCodec() {
            return OreMagnetBonus.CODEC;
        }
        
        public GemBonus validate() {
            Preconditions.checkArgument(!this.values.isEmpty(), (Object)"No values provided!");
            return this;
        }
        
        public boolean supports(final LootRarity rarity) {
            return this.values.containsKey(rarity);
        }
        
        public int getNumberOfUUIDs() {
            return 0;
        }
        
        public Component getSocketBonusTooltip(final ItemStack gem, final LootRarity rarity) {
            return (Component)Component.m_237110_("bonus." + this.getId() + ".desc", new Object[] { this.values.get(rarity).getInt(0.0f) }).m_130940_(ChatFormatting.YELLOW);
        }
        
        static {
            CODEC = RecordCodecBuilder.create(inst -> inst.group(gemClass(), (App)GemBonus.VALUES_CODEC.fieldOf("values").forGetter(a -> a.values)).apply((Applicative)inst, (BiFunction)OreMagnetBonus::new));
        }
    }
    
    public static class TreasureGoblinBonus extends GemBonus
    {
        public static final Codec<TreasureGoblinBonus> CODEC;
        protected final Map<LootRarity, Data> values;
        
        public TreasureGoblinBonus(final GemClass gemClass, final Map<LootRarity, Data> values) {
            super(Apotheosis.loc("twilight_treasure_goblin"), gemClass);
            this.values = values;
        }
        
        public void doPostAttack(final ItemStack gem, final LootRarity rarity, final LivingEntity user, final Entity target) {
            final Data d = this.values.get(rarity);
            if (Affix.isOnCooldown(this.getCooldownId(gem), d.cooldown, user)) {
                return;
            }
            if (user.f_19796_.m_188501_() <= d.chance) {
                final Redcap goblin = (Redcap)((EntityType)AdventureTwilightCompat.REDCAP.get()).m_20615_(user.m_9236_());
                final CompoundTag tag = new CompoundTag();
                tag.m_128359_("DeathLootTable", "apotheosis:entity/treasure_goblin");
                goblin.m_7378_(tag);
                goblin.getPersistentData().m_128379_("apoth.treasure_goblin", true);
                goblin.m_6593_((Component)Component.m_237115_("name.apotheosis.treasure_goblin").m_130938_(s -> s.m_131148_((TextColor)GradientColor.RAINBOW)));
                goblin.m_20340_(true);
                goblin.m_21051_(Attributes.f_22279_).m_22125_(new AttributeModifier("apoth.very_fast", 0.2, AttributeModifier.Operation.ADDITION));
                goblin.m_21051_(Attributes.f_22276_).m_22125_(new AttributeModifier("apoth.healmth", 60.0, AttributeModifier.Operation.ADDITION));
                goblin.m_21153_(goblin.m_21233_());
                for (int i = 0; i < 8; ++i) {
                    final int x = Mth.m_216271_(goblin.f_19796_, -5, 5);
                    final int y = Mth.m_216271_(goblin.f_19796_, -1, 1);
                    final int z = Mth.m_216271_(goblin.f_19796_, -5, 5);
                    goblin.m_146884_(target.m_20182_().m_82520_((double)x, (double)y, (double)z));
                    if (user.m_9236_().m_45786_((Entity)goblin)) {
                        break;
                    }
                    if (i == 7) {
                        goblin.m_146884_(target.m_20182_());
                    }
                }
                goblin.m_7292_(new MobEffectInstance(MobEffects.f_19619_, 200, 0));
                user.m_9236_().m_7967_((Entity)goblin);
                Affix.startCooldown(this.getCooldownId(gem), user);
            }
        }
        
        public Codec<? extends GemBonus> getCodec() {
            return TreasureGoblinBonus.CODEC;
        }
        
        public GemBonus validate() {
            Preconditions.checkArgument(!this.values.isEmpty(), (Object)"No values provided!");
            return this;
        }
        
        public boolean supports(final LootRarity rarity) {
            return this.values.containsKey(rarity);
        }
        
        public int getNumberOfUUIDs() {
            return 0;
        }
        
        public Component getSocketBonusTooltip(final ItemStack gem, final LootRarity rarity) {
            final Data d = this.values.get(rarity);
            final Component cooldown = (Component)Component.m_237110_("affix.apotheosis.cooldown", new Object[] { StringUtil.m_14404_(d.cooldown) });
            return (Component)Component.m_237110_("bonus." + this.getId() + ".desc", new Object[] { Affix.fmt(d.chance * 100.0f), cooldown }).m_130940_(ChatFormatting.YELLOW);
        }
        
        static {
            CODEC = RecordCodecBuilder.create(inst -> inst.group(gemClass(), (App)LootRarity.mapCodec((Codec)Data.CODEC).fieldOf("values").forGetter(a -> a.values)).apply((Applicative)inst, (BiFunction)TreasureGoblinBonus::new));
        }
        
        record Data(float chance, int cooldown) {
            public static final Codec<Data> CODEC;
            
            static {
                CODEC = RecordCodecBuilder.create(inst -> inst.group((App)Codec.FLOAT.fieldOf("chance").forGetter((Function)Data::chance), (App)Codec.INT.fieldOf("cooldown").forGetter((Function)Data::cooldown)).apply((Applicative)inst, (BiFunction)Data::new));
            }
        }
    }
    
    public static class FortificationBonus extends GemBonus
    {
        public static final Codec<FortificationBonus> CODEC;
        protected final Map<LootRarity, Data> values;
        
        public FortificationBonus(final GemClass gemClass, final Map<LootRarity, Data> values) {
            super(Apotheosis.loc("twilight_fortification"), gemClass);
            this.values = values;
        }
        
        public void doPostHurt(final ItemStack gem, final LootRarity rarity, final LivingEntity user, final Entity attacker) {
            final Data d = this.values.get(rarity);
            if (Affix.isOnCooldown(this.getCooldownId(gem), d.cooldown, user)) {
                return;
            }
            if (user.f_19796_.m_188501_() <= d.chance) {
                user.getCapability(CapabilityList.SHIELDS).ifPresent(cap -> cap.replenishShields());
                Affix.startCooldown(this.getCooldownId(gem), user);
            }
        }
        
        public Codec<? extends GemBonus> getCodec() {
            return FortificationBonus.CODEC;
        }
        
        public GemBonus validate() {
            Preconditions.checkArgument(!this.values.isEmpty(), (Object)"No values provided!");
            return this;
        }
        
        public boolean supports(final LootRarity rarity) {
            return this.values.containsKey(rarity);
        }
        
        public int getNumberOfUUIDs() {
            return 0;
        }
        
        public Component getSocketBonusTooltip(final ItemStack gem, final LootRarity rarity) {
            final Data d = this.values.get(rarity);
            final Component cooldown = (Component)Component.m_237110_("affix.apotheosis.cooldown", new Object[] { StringUtil.m_14404_(d.cooldown) });
            return (Component)Component.m_237110_("bonus." + this.getId() + ".desc", new Object[] { Affix.fmt(d.chance * 100.0f), cooldown }).m_130940_(ChatFormatting.YELLOW);
        }
        
        static {
            CODEC = RecordCodecBuilder.create(inst -> inst.group(gemClass(), (App)LootRarity.mapCodec((Codec)Data.CODEC).fieldOf("values").forGetter(a -> a.values)).apply((Applicative)inst, (BiFunction)FortificationBonus::new));
        }
        
        record Data(float chance, int cooldown) {
            public static final Codec<Data> CODEC;
            
            static {
                CODEC = RecordCodecBuilder.create(inst -> inst.group((App)Codec.FLOAT.fieldOf("chance").forGetter((Function)Data::chance), (App)Codec.INT.fieldOf("cooldown").forGetter((Function)Data::cooldown)).apply((Applicative)inst, (BiFunction)Data::new));
            }
        }
    }
}
